from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from envoy_schema.admin.schema.doe import DynamicOperatingEnvelopeResponse
from envoy_schema.admin.schema.pricing import TariffGeneratedRateResponse
from envoy_schema.admin.schema.site import SiteResponse


class ArchiveBase(BaseModel):
    """The common properties for all archive responses"""

    archive_id: int  # Unique PK for identifying an individual archive record
    archive_time: datetime  # When was this archive record created
    deleted_time: Optional[datetime]  # When was this archive record marked as deleted (if None - not deleted)


class ArchiveDynamicOperatingEnvelopeResponse(ArchiveBase, DynamicOperatingEnvelopeResponse):
    """Archived version of DynamicOperatingEnvelopeResponse"""

    pass


class ArchiveTariffGeneratedRateResponse(ArchiveBase, TariffGeneratedRateResponse):
    """Archived version of TariffGeneratedRateResponse"""

    pass


class ArchiveSiteResponse(ArchiveBase, SiteResponse):
    """Archived version of SiteResponse"""

    pass



from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema import uri
from envoy_schema.server.schema.sep2.identification import ListLink, Resource
from envoy_schema.server.schema.sep2.types import DEFAULT_POLLRATE_SECONDS


# Per the Sep2, DeviceCapability should be a subclass of FunctionSetAssignmentsBase
# However, DERProgramListUri and TariffProfileListUri are now site-scoped, which
# breaks how links are populated in DeviceCapability.
# Since clients are expected to access DERProgramList and TariffProfileList through
# FunctionSetAssignments rather than through DeviceCapability, we can disable them
# by not subclassing FunctionSetAssignmentsBase but subclassing Resource instead.
class DeviceCapabilityResponse(Resource, tag="DeviceCapability"):
    href: str = attr(default=uri.DeviceCapabilityUri)
    pollRate: Optional[int] = attr(default=DEFAULT_POLLRATE_SECONDS)  # recommended client pollrate in seconds

    # (0..1) Link
    # Not supported at this time
    # SelfDeviceLink: Optional[Link] = element(default=None)

    # (0..1) ListLink
    EndDeviceListLink: Optional[ListLink] = element(default=None)
    MirrorUsagePointListLink: Optional[ListLink] = element(default=None)



from datetime import datetime

from pydantic import BaseModel


class AggregatorDomain(BaseModel):
    """A domain whitelisted by the utility server as being controlled by the parent aggregator"""

    domain: str
    created_time: datetime
    changed_time: datetime


class AggregatorResponse(BaseModel):
    """Metadata about a specific aggregator registered in the utility server"""

    aggregator_id: int
    name: str
    domains: list[AggregatorDomain]


class AggregatorPageResponse(BaseModel):
    """Represents a paginated response of Aggregators"""

    total_count: int  # The total number of aggregators (independent of this page of results)
    limit: int  # The maximum number of aggregators that could've been returned (the limit set by the query)
    start: int  # The number of aggregators that have been skipped as part of this query (the start set by the query)
    aggregators: list[AggregatorResponse]  # The aggregator models in this page



"""Time resource related
"""

from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2.identification import Resource
from envoy_schema.server.schema.sep2.types import TimeOffsetType, TimeQualityType, TimeType


class TimeResponse(Resource, tag="Time"):
    # xsd
    href: str = attr()

    currentTime: TimeType = element()
    dstEndTime: TimeType = element()
    dstOffset: TimeOffsetType = element()
    dstStartTime: TimeType = element()
    localTime: Optional[TimeType] = element(default=None)
    quality: TimeQualityType = element()
    tzOffset: TimeOffsetType = element()



from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2 import primitive_types, types
from envoy_schema.server.schema.sep2.identification import IdentifiedObject, Link, ListLink, Resource, SubscribableList


class ReadingBase(Resource):
    consumptionBlock: Optional[types.ConsumptionBlockType] = element(default=0)
    qualityFlags: Optional[primitive_types.HexBinary16] = element(
        default=primitive_types.HexBinary16("00")
    )  # string (hex encoded) form that maps to QualityFlagsType
    timePeriod: Optional[types.DateTimeIntervalType] = element(
        default=None, tag="timePeriod"
    )  # Necessary due to DateTimeIntervalType defn
    touTier: Optional[types.TOUType] = element(default=0)
    value: Optional[int] = element(default=None)


class Reading(ReadingBase):
    localID: Optional[primitive_types.HexBinary16] = element(default=None)
    subscribable: Optional[types.SubscribableType] = attr(default=None)


class ReadingSetBase(IdentifiedObject):
    timePeriod: types.DateTimeIntervalType = element(tag="timePeriod")  # Necessary due to DateTimeIntervalType defn


class ReadingType(Resource):
    """Type of data conveyed by a specific Reading. See IEC 61968 Part 9 Annex C for full definitions
    of these values."""

    accumulationBehaviour: Optional[types.AccumulationBehaviourType] = element(default=None)
    calorificValue: Optional[types.UnitValueType] = element(default=None)
    commodity: Optional[types.CommodityType] = element(default=None)
    conversionFactor: Optional[types.UnitValueType] = element(default=None)
    dataQualifier: Optional[types.DataQualifierType] = element(default=None)
    flowDirection: Optional[types.FlowDirectionType] = element(default=None)
    intervalLength: Optional[int] = element(default=None)
    kind: Optional[types.KindType] = element(default=None)
    maxNumberOfIntervals: Optional[int] = element(default=None)
    numberOfConsumptionBlocks: Optional[int] = element(default=None)
    numberOfTouTiers: Optional[int] = element(default=None)
    phase: Optional[types.PhaseCode] = element(default=None)
    powerOfTenMultiplier: Optional[int] = element(default=None)
    subIntervalLength: Optional[int] = element(default=None)
    supplyLimit: Optional[int] = element(default=None)
    tieredConsumptionBlocks: Optional[bool] = element(default=None)
    uom: Optional[types.UomType] = element(default=None)


class UsagePointBase(IdentifiedObject):
    """Logical point on a network at which consumption or production is either physically measured (e.g. metered) or
    estimated (e.g. unmetered street lights)."""

    roleFlags: primitive_types.HexBinary16 = element()
    serviceCategoryKind: types.ServiceKind = element()
    status: int = element()


class UsagePoint(UsagePointBase):
    """Logical point on a network at which consumption or production is either physically measured (e.g. metered) or
    estimated (e.g. unmetered street lights)."""

    deviceLFDI: str = element()
    MeterReadingListLink: Optional[ListLink] = element(default=None)


class MeterReading(IdentifiedObject):
    """Set of values obtained from the meter."""

    RateComponentListLink: Optional[ListLink] = element(default=None)
    ReadingLink: Optional[Link] = element(default=None)
    ReadingSetListLink: Optional[ListLink] = element(default=None)
    ReadingTypeLink: Link = element()


class ReadingSet(ReadingSetBase):
    """A set of Readings of the ReadingType indicated by the parent MeterReading."""

    ReadingListLink: Optional[ListLink] = element(default=None)


class ReadingListResponse(SubscribableList, tag="ReadingList"):
    Readings: Optional[list["Reading"]] = element(default=None, tag="Reading")



# Defines all the URIs
# Some URIs have named parameters e.g. EndDeviceUri = "/edev/{site_id}". These are intended to be filled in with an
# appropriate format statement, for example, EndDeviceUri.format(site_id=4) or EndDeviceUri
# Note: 'id1', 'id2', 'id3' are placeholders and should be replaced with suitable names matching their purpose.

# CSIP-Aus URIs
ConnectionPointUri = "/edev/{site_id}/cp"

# Sep2 URIs
AccountBalanceUri = "/ppy/{id1}/ab"
ActiveBillingPeriodListUri = "/bill/{id1}/ca/{id2}/actbp"
ActiveCreditRegisterListUri = ""  # noqa: E501. There is NO ActiveCreditRegisterList resource in Sep2 despite an ActiveCreditRegisterListLink being defined.
ActiveDERControlListUri = "/edev/{site_id}/derp/{der_program_id}/actderc"
ActiveEndDeviceControlListUri = "/dr/{id1}/actedc"
ActiveFlowReservationListUri = ""  # noqa: E501. There is NO ActiveFlowReservationList resource in Sep2 despite an ActiveFlowReservationListLink being defined.
ActiveProjectionReadingListUri = ""  # noqa: E501. There is NO ActiveProjectionReadingList in resource Sep2 despite an ActiveProjectionReadingListLink being defined.
ActiveSupplyInterruptionOverrideListUri = "/ppy/{id1}/actsi"
ActiveTargetReadingListUri = ""  # noqa: E501. There is NO ActiveTargetReadingList resource in Sep2 despite an ActiveTargetReadingListLink being defined.
ActiveTextMessageListUri = "/msg/{id1}/acttxt"
ActiveTimeTariffIntervalListUri = "/tp/{id1}/rc/{id2}/acttti"
AssociatedDERProgramListUri = "/edev/{site_id}/der/{der_id}/derp"
AssociatedUsagePointUri = "/edev/{site_id}/der/{der_id}/upt"
BillingPeriodListUri = "/bill/{id1}/ca/{id2}/bp"
BillingReadingListUri = "/brs/{id1}/br"
BillingReadingSetListUri = "/brs"
ConfigurationUri = "/edev/{site_id}/cfg"
ConsumptionTariffIntervalListUri = (
    "/edev/{site_id}/tp/{tariff_id}/rc/{rate_component_id}/{pricing_reading}/tti/{tti_id}/cti/{sep2_price}"
)
ConsumptionTariffIntervalUri = (
    "/edev/{site_id}/tp/{tariff_id}/rc/{rate_component_id}/{pricing_reading}/tti/{tti_id}/cti/{sep2_price}/1"
)
CreditRegisterListUri = "/ppy/{id1}/cr"
CustomerAccountListUri = "/bill"
CustomerAccountUri = "/bill/{id1}"
CustomerAgreementListUri = "/bill/{id1}/ca"
DefaultDERControlUri = "/edev/{site_id}/derp/{der_program_id}/dderc"
DemandResponseProgramListUri = "/dr"  # /drp used in Step 11 p275 of Sep2
DemandResponseProgramUri = "/dr/{id1}"
DERAvailabilityUri = "/edev/{site_id}/der/{der_id}/dera"
DERCapabilityUri = "/edev/{site_id}/der/{der_id}/dercap"
DERControlListUri = "/edev/{site_id}/derp/{der_program_id}/derc"
DERControlAndListByDateUri = "/edev/{site_id}/derp/{der_program_id}/derc/{derc_id_or_date}"
DERCurveListUri = "/derp/{id1}/dc"
DERCurveUri = "/derp/{id1}/dc/{id2}"
DERListUri = "/edev/{site_id}/der"
# DERProgramListUri = "/derp"
# DERProgramUri = "/derp/{id1}"
# Modify DERProgram URIs to be site-scoped
DERProgramListUri = "/edev/{site_id}/derp"
DERProgramUri = "/edev/{site_id}/derp/{der_program_id}"
DERSettingsUri = "/edev/{site_id}/der/{der_id}/derg"
DERStatusUri = "/edev/{site_id}/der/{der_id}/ders"
DERUri = "/edev/{site_id}/der/{der_id}"
DeviceCapabilityUri = "/dcap"
DeviceInformationUri = "/edev/{site_id}/di"
DeviceStatusUri = "/edev/{site_id}/dstat"
EndDeviceControlListUri = "/dr/{id1}/edc"
EndDeviceListUri = "/edev"
EndDeviceUri = "/edev/{site_id}"
FileListUri = "/file"
FileStatusUri = "/edev/{site_id}/fs"
FileUri = "/file/{id1}"
FlowReservationRequestListUri = "/edev/{site_id}/frq"
FlowReservationResponseListUri = "/edev/{site_id}/frp"
FunctionSetAssignmentsListUri = "/edev/{site_id}/fsa"
FunctionSetAssignmentsUri = "/edev/{site_id}/fsa/{fsa_id}"
HistoricalReadingListUri = "/bill/{id1}/ca/{id2}/ver"
IPAddrListUri = "/edev/{site_id}/ns/{id2}/addr"
IPInterfaceListUri = "/edev/{site_id}/ns"
LLInterfaceListUri = "/edev/{site_id}/ns/{id2}/ll"
LoadShedAvailabilityListUri = "/edev/{site_id}/lsl"
LogEventListUri = "/edev/{site_id}/lel"
MessagingProgramListUri = "/msg"
MeterReadingListUri = "/upt/{id1}/mr"
MeterReadingUri = "/upt/{id1}/mr/{id2}"
MirrorUsagePointListUri = "/mup"
MirrorUsagePointUri = "/mup/{mup_id}"
NeighborListUri = "/edev/{site_id}/ns/{id2}/ll/{id3}/nbh"
NotificationListUri = "/ntfy"
PowerStatusUri = "/edev/{site_id}/ps"
PrepaymentListUri = "/ppy"
PrepaymentUri = "/ppy/{id1}"
PrepayOperationStatusUri = "/ppy/{id1}/os"
PriceResponseCfgListUri = "/edev/{site_id}/cfg/prcfg"
PricingReadingTypeUri = "/pricing/rt/{reading_type}"
ProjectionReadingListUri = "/bill/{id1}/ca/{id2}/pro"
RateComponentListUnscopedUri = "/tp/{tariff_id}/rc"
RateComponentListUri = "/edev/{site_id}/tp/{tariff_id}/rc"
RateComponentUri = "/edev/{site_id}/tp/{tariff_id}/rc/{rate_component_id}/{pricing_reading}"
ReadingListUri = "/upt/{site_id}/mr/{site_reading_type_id}/rs/{reading_set_id}/r"
ReadingSetListUri = "/upt/{id1}/mr/{id2}/rs"
ReadingTypeUri = "/upt/{id1}/mr/{id2}/rt"
ReadingUri = "/upt/{id1}/mr/{id2}/rs/{id3}/r/{id4}"
RegistrationUri = "/edev/{site_id}/rg"
ResponseListUri = "/rsps/{id1}/rsp"
ResponseSetListUri = "/rsps"
RPLInstanceListUri = "/edev/{site_id}/ns/{id2}/addr/{id3}/rpl"
RPLSourceRoutesListUri = "/edev/{site_id}/ns/{id2}/addr/{id3}/rpl/{id4}/srt"
SelfDeviceUri = "/sdev"
ServiceSupplierUri = "/bill/{id1}/ss"
SubscriptionGlobalUri = "/sub/{subscription_id}"  # This is outside the standard but is necessary for some subscriptions
SubscriptionListGlobalUri = "/sub"  # This is outside the standard but is necessary for some subscriptions
SubscriptionListUri = "/edev/{site_id}/sub"
SubscriptionUri = "/edev/{site_id}/sub/{subscription_id}"
SupplyInterruptionOverrideListUri = "/ppy/{id1}/si"
SupportedLocaleListUri = "/edev/{site_id}/di/loc"
TargetReadingListUri = "/bill/{id1}/ca/{id2}/tar"
TariffProfileListUnscopedUri = "/tp"
TariffProfileListUri = "/edev/{site_id}/tp"
TariffProfileUnscopedUri = "/tp/{tariff_id}"
TariffProfileUri = "/edev/{site_id}/tp/{tariff_id}"
TextMessageListUri = "/msg/{id1}/txt"
TimeTariffIntervalListUri = "/edev/{site_id}/tp/{tariff_id}/rc/{rate_component_id}/{pricing_reading}/tti"
TimeTariffIntervalUri = "/edev/{site_id}/tp/{tariff_id}/rc/{rate_component_id}/{pricing_reading}/tti/{tti_id}"
TimeUri = "/tm"
UsagePointListUri = "/upt"
UsagePointUri = "/upt/{id1}"



from typing import Optional

from pydantic_xml import element

from envoy_schema.server.schema.sep2.identification import RespondableSubscribableIdentifiedObject
from envoy_schema.server.schema.sep2.types import DateTimeIntervalType, OneHourRangeType, TimeType
from envoy_schema.server.schema.sep2.primitive_types import String192
from envoy_schema.server.schema.sep2.base import BaseXmlModelWithNS


class EventStatus(BaseXmlModelWithNS):
    """Current status information relevant to a specific object. The Status object is used to indicate the current
    status of an Event. Devices can read the containing resource (e.g. TextMessage) to get the most up to date status of
    the event.  Devices can also subscribe to a specific resource instance to get updates when any of its attributes
    change, including the Status object."""

    currentStatus: int = element()
    dateTime: TimeType = element()
    potentiallySuperseded: bool = element()
    potentiallySupersededTime: Optional[TimeType] = element(default=None)
    reason: Optional[String192] = element(default=None)


class Event(RespondableSubscribableIdentifiedObject):
    """An Event indicates information that applies to a particular period of time. Events SHALL be executed relative
    to the time of the server, as described in the Time function set section 11.1."""

    creationTime: TimeType = element()
    EventStatus_: EventStatus = element(tag="EventStatus")
    interval: DateTimeIntervalType = element(tag="interval")


class RandomizableEvent(Event):
    """An Event that can indicate time ranges over which the start time and duration SHALL be randomized."""

    randomizeDuration: Optional[OneHourRangeType] = element(default=None)
    randomizeStart: Optional[OneHourRangeType] = element(default=None)



from enum import IntEnum, IntFlag, auto
from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2 import primitive_types, types
from envoy_schema.server.schema.sep2.base import BaseXmlModelWithNS
from envoy_schema.server.schema.sep2.der_control_types import (
    ActivePower,
    AmpereHour,
    ApparentPower,
    CurrentRMS,
    FixedVar,
    PowerFactor,
    PowerFactorWithExcitation,
    ReactivePower,
    ReactiveSusceptance,
    VoltageRMS,
    WattHour,
)
from envoy_schema.server.schema.sep2.event import RandomizableEvent
from envoy_schema.server.schema.sep2.identification import IdentifiedObject, Link
from envoy_schema.server.schema.sep2.identification import List
from envoy_schema.server.schema.sep2.identification import List as Sep2List
from envoy_schema.server.schema.sep2.identification import (
    ListLink,
    SubscribableIdentifiedObject,
    SubscribableList,
    SubscribableResource,
)
from envoy_schema.server.schema.sep2.pricing import PrimacyType


class DERType(IntEnum):
    NOT_APPLICABLE = 0
    VIRTUAL_OR_MIXED = 1
    RECIPROCATING_ENGINE = 2
    FUEL_CELL = 3
    PHOTOVOLTAIC_SYSTEM = 4
    COMBINED_HEAT_POWER = 5
    OTHER_GENERATION_SYSTEM = 6
    OTHER_STORAGE_SYSTEM = 80
    ELECTRIC_VEHICLE = 81
    EVSE = 82
    COMBINED_PV_AND_STORAGE = 83


class DERControlType(IntFlag):
    """Series of bit flags: Control modes supported by the DER"""

    CHARGE_MODE = auto()
    DISCHARGE_MODE = auto()
    OP_MOD_CONNECT = auto()  # Connect/Disconnect - implies galvanic isolation
    OP_MOD_ENERGIZE = auto()  # Energize/De-Energize
    OP_MOD_FIXED_PF_ABSORB_W = auto()  # Fixed Power Factor Setpoint when absorbing active power
    OP_MOD_FIXED_PF_INJECT_W = auto()  # Fixed Power Factor Setpoint when injecting active power
    OP_MOD_FIXED_VAR = auto()  # Reactive power setpoint
    OP_MOD_FIXED_W = auto()  # Charge / Discharge Setpoint
    OP_MOD_FREQ_DROOP = auto()  # Frequency-Watt Parameterized Mode
    OP_MOD_FREQ_WATT = auto()  # Frequency-Watt Curve Mode
    OP_MOD_HFRT_MAY_TRIP = auto()  # High Frequency Ride Through, May Trip Mode
    OP_MOD_HFRT_MUST_TRIP = auto()  # High Frequency Ride Through, Must Trip Mode
    OP_MOD_HVRT_MAY_TRIP = auto()  # High Voltage Ride Through, May Trip Mode
    OP_MOD_HVRT_MOMENTARY_CESSATION = auto()  # High Frequency Ride Through, Momentary cessation Mode
    OP_MOD_HVRT_MUST_TRIP = auto()  # High Voltage Ride Through, Must Trip Mode
    OP_MOD_LFRT_MAY_TRIP = auto()  # Low Frequency Ride Through, May Trip Mode
    OP_MOD_LFRT_MUST_TRIP = auto()  # Low Frequency Ride Through, Must Trip Mode
    OP_MOD_LVRT_MAY_TRIP = auto()  # Low Voltage Ride Through, May Trip Mode
    OP_MOD_LVRT_MOMENTARY_CESSATION = auto()  # Low Frequency Ride Through, Momentary cessation Mode
    OP_MOD_LVRT_MUST_TRIP = auto()  # Low Voltage Ride Through, Must Trip Mode
    OP_MOD_MAX_LIM_W = auto()  # Maximum Active Power
    OP_MOD_TARGET_VAR = auto()  # Target Reactive Power
    OP_MOD_TARGET_W = auto()  # Target Active Power
    OP_MOD_VOLT_VAR = auto()  # Volt-Var Mode
    OP_MOD_VOLT_WATT = auto()  # Volt-Watt Mode
    OP_MOD_WATT_PF = auto()  # Watt-PowerFactor Mode
    OP_MOD_WATT_VAR = auto()  # Watt-Var Mode


class InverterStatusType(IntEnum):
    """DER InverterStatus value"""

    NOT_APPLICABLE = 0
    OFF = 1
    SLEEPING = 2  # sleeping (auto-shutdown) or DER is at low output power/voltage
    STARTING = 3  # starting up or ON but not producing power
    TRACKING_MPPT_POWER_POINT = 4  # tracking MPPT power point
    FORCED_POWER_REDUCTION = 5  # forced power reduction/derating
    SHUTTING_DOWN = 6
    ONE_OR_MORE_FAULTS = 7
    STANDBY = 8  # standby (service on unit) - DER may be at high output voltage/power
    TEST_MODE = 9
    MANUFACTURER_STATUS = 10  # as defined in manufacturer status


class OperationalModeStatusType(IntEnum):
    """DER OperationalModeStatus value"""

    NOT_APPLICABLE = 0
    OFF = 1
    OPERATIONAL_MODEL = 2
    TEST_MODE = 3


class StorageModeStatusType(IntEnum):
    """DER StorageModeStatus value"""

    STORAGE_CHARGING = 0
    STORAGE_DISCHARGING = 1
    STORAGE_HOLDING = 2


class LocalControlModeStatusType(IntEnum):
    """DER LocalControlModeStatus/value"""

    LOCAL_CONTROL = 0
    REMOTE_CONTROL = 1


class ConnectStatusType(IntFlag):
    """Bit map of DER ConnectStatus values"""

    CONNECTED = auto()
    AVAILABLE = auto()
    OPERATING = auto()
    TEST = auto()
    FAULT_ERROR = auto()


class AlarmStatusType(IntFlag):
    """Bitmap indicating the status of DER alarms (see DER LogEvents for more details)."""

    DER_FAULT_OVER_CURRENT = auto()
    DER_FAULT_OVER_VOLTAGE = auto()
    DER_FAULT_UNDER_VOLTAGE = auto()
    DER_FAULT_OVER_FREQUENCY = auto()
    DER_FAULT_UNDER_FREQUENCY = auto()
    DER_FAULT_VOLTAGE_IMBALANCE = auto()
    DER_FAULT_CURRENT_IMBALANCE = auto()
    DER_FAULT_EMERGENCY_LOCAL = auto()
    DER_FAULT_EMERGENCY_REMOTE = auto()
    DER_FAULT_LOW_POWER_INPUT = auto()
    DER_FAULT_PHASE_ROTATION = auto()


class AbnormalCategoryType(IntEnum):
    """Abnormal operating performance category as defined by IEEE 1547-2018. One of:"""

    NOT_SPECIFIED = 0
    CATEGORY_1 = 1
    CATEGORY_2 = 2
    CATEGORY_3 = 3


class NormalCategoryType(IntEnum):
    """Normal operating performance category as defined by IEEE 1547-2018."""

    NOT_SPECIFIED = 0
    CATEGORY_A = 1
    CATEGORY_B = 2


class DOESupportedMode(IntFlag):
    """Series of bit flags: What CSIP Aus DOE capabilities are enabled"""

    OP_MOD_EXPORT_LIMIT_W = auto()
    OP_MOD_IMPORT_LIMIT_W = auto()
    OP_MOD_GENERATION_LIMIT_W = auto()
    OP_MOD_LOAD_LIMIT_W = auto()


class FreqDroopType(BaseXmlModelWithNS):
    """Type for Frequency-Droop (Frequency-Watt) operation."""

    dBOF: int = element()  # Frequency droop dead band for over-frequency conditions.
    dBUF: int = element()  # Frequency droop dead band for under-frequency conditions.
    kOF: int = element()  # droop per-unit frequency change OF conditions corresponding to 1 power output change.
    kUF: int = element()  # # droop per-unit frequency change UF conditions corresponding to 1 power output change.
    openLoopTms: int = element()  # Open loop response time


class DERControlBase(BaseXmlModelWithNS):
    """Distributed Energy Resource (DER) control values."""

    opModConnect: Optional[bool] = element(default=None)  # Set DER as connected (true) or disconnected (false).
    opModEnergize: Optional[bool] = element(default=None)  # Set DER as energized (true) or de-energized (false).
    opModFixedPFAbsorbW: Optional[PowerFactorWithExcitation] = element(
        default=None
    )  # requested PF when AP is being absorbed
    opModFixedPFInjectW: Optional[PowerFactorWithExcitation] = element(
        default=None
    )  # requested PF when AP is being injected
    opModFixedVar: Optional[FixedVar] = element(default=None)  # specifies the delivered or received RP setpoint.
    opModFixedW: Optional[types.SignedPerCent] = element(
        default=None
    )  # specifies a requested charge/discharge mode setpoint
    opModFreqDroop: Optional[FreqDroopType] = element(default=None)  # Specifies a frequency-watt operation
    opModFreqWatt: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 0
    opModHFRTMayTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 1
    opModHFRTMustTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 2
    opModHVRTMayTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 3
    opModHVRTMomentaryCessation: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 4
    opModHVRTMustTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 5
    opModLFRTMayTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 6
    opModLFRTMustTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 7
    opModLVRTMayTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 8
    opModLVRTMomentaryCessation: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 9
    opModLVRTMustTrip: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 10
    opModMaxLimW: Optional[types.PerCent] = element(
        default=None
    )  # max active power generation level at electrical coupling point
    opModTargetVar: Optional[ReactivePower] = element(default=None)  # Target reactive power, in var
    opModTargetW: Optional[ActivePower] = element(default=None)  # Target active power, in Watts
    opModVoltVar: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 11
    opModVoltWatt: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 12
    opModWattPF: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 13
    opModWattVar: Optional[Link] = element(default=None)  # Specify DERCurveLink for curveType == 14
    rampTms: Optional[int] = element(default=None)  # Requested ramp time, in hundredths of a second

    # CSIP Aus Extensions (encoded here as it makes decoding a whole lot simpler)
    opModImpLimW: Optional[ActivePower] = element(
        ns="csipaus", default=None
    )  # constraint on the imported AP at the connection point
    opModExpLimW: Optional[ActivePower] = element(
        ns="csipaus", default=None
    )  # constraint on the exported AP at the connection point
    opModGenLimW: Optional[ActivePower] = element(
        ns="csipaus", default=None
    )  # max limit on discharge watts for a single DER
    opModLoadLimW: Optional[ActivePower] = element(
        ns="csipaus", default=None
    )  # max limit on charge watts for a single DER


class DefaultDERControl(SubscribableIdentifiedObject):
    """Contains control mode information to be used if no active DERControl is found."""

    DERControlBase_: DERControlBase = element(tag="DERControlBase")
    setESDelay: Optional[int] = element(default=None)  # Enter service delay, in hundredths of a second.
    setESHighFreq: Optional[int] = element(default=None)  # Enter service frequency high. Specified in hundredths of Hz
    setESHighVolt: Optional[int] = element(
        default=None
    )  # Enter service voltage high. Specified as an effective percent voltage,
    setESLowFreq: Optional[int] = element(default=None)  # Enter service frequency low. Specified in hundredths of Hz
    setESLowVolt: Optional[int] = element(
        default=None
    )  # Enter service voltage low. Specified as an effective percent voltage,
    setESRampTms: Optional[int] = element(default=None)  # Enter service ramp time, in hundredths of a second
    setESRandomDelay: Optional[int] = element(
        default=None
    )  # Enter service randomized delay, in hundredths of a second.
    setGradW: Optional[int] = element(default=None)  # Set default rate of change (ramp rate) of active power output
    setSoftGradW: Optional[int] = element(
        default=None
    )  # Set soft-start rate of change (soft-start ramp rate) of AP output


class DERControlResponse(RandomizableEvent, tag="DERControl"):
    """Distributed Energy Resource (DER) time/event-based control."""

    DERControlBase_: DERControlBase = element(tag="DERControlBase")
    deviceCategory: Optional[primitive_types.HexBinary32] = element(
        default=None,
    )  # the bitmap indicating device categories that SHOULD respond.


class DERControlListResponse(SubscribableList, tag="DERControlList"):
    DERControl: Optional[list[DERControlResponse]] = element(default=None)


class DERProgramResponse(SubscribableIdentifiedObject, tag="DERProgram"):
    """sep2 DERProgram"""

    ActiveDERControlListLink: Optional[ListLink] = element(default=None)
    DefaultDERControlLink: Optional[Link] = element(default=None)
    DERControlListLink: Optional[ListLink] = element(default=None)
    DERCurveListLink: Optional[ListLink] = element(default=None)
    primacy: PrimacyType = element()


class DERProgramListResponse(SubscribableList, tag="DERProgramList"):
    DERProgram: Optional[list[DERProgramResponse]] = element(default=None)
    pollRate: Optional[int] = attr(
        default=None
    )  # The default polling rate for this resource and all resources below in seconds


class DemandResponseProgramResponse(IdentifiedObject, tag="DemandResponseProgram"):
    """sep2 Demand response program"""

    ActiveEndDeviceControlListLink: Optional[ListLink] = element(default=None)
    availabilityUpdatePercentChangeThreshold: Optional[types.PerCent] = element(default=None)
    availabilityUpdatePowerChangeThreshold: Optional[ActivePower] = element(default=None)
    EndDeviceControlListLink: Optional[ListLink] = element(default=None)
    primacy: PrimacyType = element()


class DemandResponseProgramListResponse(Sep2List, tag="DemandResponseProgramList"):
    DemandResponseProgram: list[DemandResponseProgramResponse] = element(default_factory=list)


class EndDeviceControlResponse(RandomizableEvent, tag="EndDeviceControl"):
    """Instructs an EndDevice to perform a specified action."""

    deviceCategory: primitive_types.HexBinary32 = element()  # HexBinary Encoded types.DeviceCategory enum
    drProgramMandatory: bool = element()
    loadShiftForward: bool = element()
    overrideDuration: Optional[int] = element(default=None)


class DER(SubscribableResource):
    """sep2 DER: Contains links to DER resources."""

    AssociatedDERProgramListLink: Optional[ListLink] = element(
        default=None
    )  # Link to List of DERPrograms having the DERControls for this DER
    AssociatedUsagePointLink: Optional[Link] = element(
        default=None
    )  # If present, this is the submeter that monitors the DER output.

    CurrentDERProgramLink: Optional[Link] = element(
        default=None
    )  # If set, this is the DERProgram containing the currently active DERControl

    DERAvailabilityLink: Optional[Link] = element(
        default=None
    )  # SHALL contain a Link to an instance of DERAvailability.
    DERCapabilityLink: Optional[Link] = element(default=None)  # SHALL contain a Link to an instance of DERCapability.
    DERSettingsLink: Optional[Link] = element(default=None)  # SHALL contain a Link to an instance of DERSettings.

    DERStatusLink: Optional[Link] = element(default=None)  # SHALL contain a Link to an instance of DERStatus.


class ConnectStatusTypeValue(BaseXmlModelWithNS, tag="ConnectStatusType"):
    dateTime: types.TimeType = element()  # The date and time at which the state applied.
    value: primitive_types.HexBinary8 = element()  # Should have bits set from ConnectStatusType


class InverterStatusTypeValue(BaseXmlModelWithNS, tag="InverterStatusType"):
    dateTime: types.TimeType = element()  # The date and time at which the state applied.
    value: InverterStatusType = element()


class LocalControlModeStatusTypeValue(BaseXmlModelWithNS, tag="LocalControlModeStatusType"):
    dateTime: types.TimeType = element()  # The date and time at which the state applied.
    value: LocalControlModeStatusType = element()


class OperationalModeStatusTypeValue(BaseXmlModelWithNS, tag="OperationalModeStatusType"):
    dateTime: types.TimeType = element()  # The date and time at which the state applied.
    value: OperationalModeStatusType = element()


class StorageModeStatusTypeValue(BaseXmlModelWithNS, tag="StorageModeStatusType"):
    dateTime: types.TimeType = element()  # The date and time at which the state applied.
    value: StorageModeStatusType = element()


class ManufacturerStatusValue(BaseXmlModelWithNS, tag="ManufacturerStatusType"):
    dateTime: types.TimeType = element()  # The date and time at which the state applied.
    value: primitive_types.String6 = element()  # The manufacturer status value


class StateOfChargeStatusValue(BaseXmlModelWithNS, tag="StateOfChargeStatusType"):
    dateTime: types.TimeType = element()  # The date and time at which the state applied.
    value: types.PerCent = element()


class DERStatus(SubscribableResource):
    """DER status information"""

    # Pydantic looks for tags in subclasses if one isnt explicitly defined. Hence redundant tags are placed here,
    # e.g. genConnectStatus, otherwise they will be renamed. Removing the tags in the subclasses would remove them
    # from xsd validation

    alarmStatus: Optional[primitive_types.HexBinary32] = element(default=None)  # AlarmStatusType encoded HexBinary str
    genConnectStatus: Optional[ConnectStatusTypeValue] = element(
        default=None, tag="genConnectStatus"
    )  # Connection status for generator
    inverterStatus: Optional[InverterStatusTypeValue] = element(default=None, tag="inverterStatus")
    localControlModeStatus: Optional[LocalControlModeStatusTypeValue] = element(
        default=None, tag="localControlModeStatus"
    )
    manufacturerStatus: Optional[ManufacturerStatusValue] = element(default=None, tag="manufacturerStatus")
    operationalModeStatus: Optional[OperationalModeStatusTypeValue] = element(default=None, tag="operationalModeStatus")
    readingTime: types.TimeType = element()
    stateOfChargeStatus: Optional[StateOfChargeStatusValue] = element(default=None, tag="stateOfChargeStatus")
    storageModeStatus: Optional[StorageModeStatusTypeValue] = element(default=None, tag="storageModeStatus")
    storConnectStatus: Optional[ConnectStatusTypeValue] = element(
        default=None, tag="storConnectStatus"
    )  # Connection status for storage


class DERAvailability(SubscribableResource):
    """Indicates current reserve generation status"""

    availabilityDuration: Optional[int] = element(
        default=None
    )  # Indicates number of seconds the DER will be able to deliver active power at the reservePercent level.
    maxChargeDuration: Optional[int] = element(
        default=None
    )  # Indicates number of seconds the DER will be able to receive active power at the reserveChargePercent level.
    readingTime: types.TimeType = element()  # The timestamp when the DER availability was last updated.
    reserveChargePercent: Optional[types.PerCent] = element(
        default=None
    )  # Percent of continuous received active power (%setMaxChargeRateW) that is estimated to be available in reserve.
    reservePercent: Optional[types.PerCent] = element(
        default=None
    )  # Percent of continuous delivered active power (%setMaxW) that is estimated to be available in reserve.
    statVarAvail: Optional[ReactivePower] = element(
        default=None
    )  # Estimated reserve reactive power, in var.  Represents the lesser of received or delivered reactive power.
    statWAvail: Optional[ActivePower] = element(default=None)  # Estimated reserve active power, in watts.


class DERCapability(SubscribableResource):
    """Distributed energy resource type and nameplate ratings. Intentionally differs from sep which is defined with a
    Resource base class rather than SubscribableResource."""

    modesSupported: primitive_types.HexBinary32 = element()  # HexBinary encoded DERControlType flags
    rtgAbnormalCategory: Optional[AbnormalCategoryType] = element(default=None)  #
    rtgMaxA: Optional[CurrentRMS] = element(default=None)  # Maximum continuous AC current capability of the DER
    rtgMaxAh: Optional[AmpereHour] = element(default=None)  # Usable energy storage capacity of the DER, in AmpHours.
    rtgMaxChargeRateVA: Optional[ApparentPower] = element(
        default=None
    )  # Maximum apparent power charge rating in Volt-Amperes. May differ from the maximum apparent power rating.
    rtgMaxChargeRateW: Optional[ActivePower] = element(
        default=None
    )  # Maximum rate of energy transfer received by the storage DER, in Watts.
    rtgMaxDischargeRateVA: Optional[ApparentPower] = element(
        default=None
    )  # Maximum apparent power discharge rating in Volt-Amperes. May differ from the maximum apparent power rating.
    rtgMaxDischargeRateW: Optional[ActivePower] = element(
        default=None
    )  # Maximum rate of energy transfer delivered by the storage DER, in Watts. Required for DERType == 83
    rtgMaxV: Optional[VoltageRMS] = element(default=None)  # AC voltage maximum rating.
    rtgMaxVA: Optional[ApparentPower] = element(default=None)  # Maximum cont' apparent power output capability, in VA.
    rtgMaxVar: Optional[ReactivePower] = element(default=None)  # Max cont' reactive power delivered by the DER (var)
    rtgMaxVarNeg: Optional[ReactivePower] = element(
        default=None
    )  # Maximum continuous reactive power received by the DER, in var. If absent, defaults to negative rtgMaxVar.
    rtgMaxW: ActivePower = (
        element()
    )  # Max cont' AP output capability of the DER, in watts. Combined gen plus storage output if DERType == 83.
    rtgMaxWh: Optional[WattHour] = element(default=None)  # Maximum energy storage capacity of the DER, in WattHours.
    rtgMinPFOverExcited: Optional[PowerFactor] = element(
        default=None
    )  # Minimum Power Factor displacement capability of the DER when injecting reactive power (over-excited)
    # SHALL be a positive value between 0.0 (typically > 0.7) and 1.0. If absent, defaults to unity.
    rtgMinPFUnderExcited: Optional[PowerFactor] = element(
        default=None
    )  # Minimum Power Factor displacement capability of the DER when absorbing reactive power (under-excited);
    # SHALL be a positive value between 0.0 (typically > 0.7) and 0.9999.  If absent, defaults to rtgMinPFOverExcited.
    rtgMinV: Optional[VoltageRMS] = element(default=None)  # AC voltage minimum rating.
    rtgNormalCategory: Optional[NormalCategoryType] = element(default=None)
    rtgOverExcitedPF: Optional[PowerFactor] = element(default=None)  # Specified over-excited power factor.
    rtgOverExcitedW: Optional[ActivePower] = element(
        default=None
    )  # AP rating in Watts at specified over-excited PF (rtgOverExcitedPF). If set rtgOverExcitedPF SHALL be present.
    rtgReactiveSusceptance: Optional[ReactiveSusceptance] = element(
        default=None
    )  # Reactive susceptance that remains connected to the Area EPS in the cease to energize and trip state.
    rtgUnderExcitedPF: Optional[PowerFactor] = element(default=None)  # Specified under-excited power factor.
    rtgUnderExcitedW: Optional[ActivePower] = element(default=None)  # Specified under-excited power factor.

    rtgVNom: Optional[VoltageRMS] = element(default=None)  # AC voltage nominal rating.
    type_: DERType = element(tag="type")  # Type of DER; see DERType object

    # CSIP Aus Extensions (encoded here as it makes decoding a whole lot simpler)
    # This is an encoded version of DOESupportedMode
    doeModesSupported: primitive_types.HexBinary8 = element(ns="csipaus", default=None)


class DERSettings(SubscribableResource):
    """Distributed energy resource settings"""

    modesEnabled: Optional[primitive_types.HexBinary32] = element(default=None)  # Hex encoded DERControlType flags
    setESDelay: Optional[int] = element(default=None)  # Enter service delay, in hundredths of a second.
    setESHighFreq: Optional[int] = element(default=None)  # Enter service frequency high. Specified in hundredths of Hz.
    setESHighVolt: Optional[int] = element(
        default=None
    )  # Enter service voltage high. Specified as an effective percent voltage in hundredths of a percent.
    setESLowFreq: Optional[int] = element(default=None)  # Enter service frequency low. Specified in hundredths of Hz.
    setESLowVolt: Optional[int] = element(
        default=None
    )  # Enter service voltage low. Specified as an effective percent voltage in hundredths of a percent.
    setESRampTms: Optional[int] = element(default=None)  # Enter service ramp time, in hundredths of a second.
    setESRandomDelay: Optional[int] = element(
        default=None
    )  # Enter service randomized delay, in hundredths of a second.
    setGradW: int = (
        element()
    )  # Set default rate of change (ramp rate) of active power output due to command or internal action,
    # defined in %setWMax / second.  Resolution is in hundredths of a percent/second.
    # A value of 0 means there is no limit. Interpreted as a percentage change in output capability limit
    # per second when used as a default ramp rate.
    setMaxA: Optional[CurrentRMS] = element(default=None)  # AC current maximum. Maximum AC current in RMS Amperes.
    setMaxAh: Optional[AmpereHour] = element(
        default=None
    )  # Max usable energy storage capacity of the DER, in AmpHours. This may be different from physical capability.
    setMaxChargeRateVA: Optional[ApparentPower] = element(
        default=None
    )  # Apparent power charge maximum. Maximum apparent power the DER can absorb from the grid in Volt-Amperes.
    # May differ from the apparent power maximum (setMaxVA).
    setMaxChargeRateW: Optional[ActivePower] = element(
        default=None
    )  # Maximum rate of energy transfer received by the storage device, in Watts. Defaults to rtgMaxChargeRateW.
    setMaxDischargeRateVA: Optional[ApparentPower] = element(
        default=None
    )  # Apparent power discharge maximum. Maximum apparent power the DER can deliver to the grid in Volt-Amperes.
    # May differ from the apparent power maximum (setMaxVA).
    setMaxDischargeRateW: Optional[ActivePower] = element(
        default=None
    )  # Maximum rate of energy transfer delivered by the storage device, in Watts. Defaults to rtgMaxDischargeRateW.
    setMaxV: Optional[VoltageRMS] = element(default=None)  # AC voltage maximum setting.
    setMaxVar: Optional[ReactivePower] = element(
        default=None
    )  # Set limit for maximum apparent power capability of the DER (in VA). Defaults to rtgMaxVA.
    setMaxVarNeg: Optional[ReactivePower] = element(
        default=None
    )  # Set limit for maximum reactive power received by the DER (in var).
    # If present, SHALL be a negative value >= rtgMaxVarNeg (default). If absent, defaults to negative setMaxVar.
    setMaxW: ActivePower = (
        element()
    )  # Set limit for maximum active power capability of the DER (in W). Defaults to rtgMaxW.
    setMaxWh: Optional[WattHour] = element(
        default=None
    )  # Maximum energy storage capacity of the DER, in WattHours. Note: this may be different from physical capability.
    setMinPFOverExcited: Optional[PowerFactor] = element(
        default=None
    )  # Set minimum Power Factor displacement limit of the DER when injecting reactive power (over-excited);
    # SHALL be a positive value between 0.0 (typically > 0.7) and 1.0.  SHALL be >= rtgMinPFOverExcited (default).
    setMinPFUnderExcited: Optional[PowerFactor] = element(
        default=None
    )  # Set minimum Power Factor displacement limit of the DER when absorbing reactive power (under-excited);
    # SHALL be a positive value between 0.0 (typically > 0.7) and 0.9999.
    # If present, SHALL be >= rtgMinPFUnderExcited (default).  If absent, defaults to setMinPFOverExcited.
    setMinV: Optional[VoltageRMS] = element(default=None)  # AC voltage min setting.
    setSoftGradW: Optional[int] = element(
        default=None
    )  # Set soft-start rate of change (soft-start ramp rate) of active power output due to command or internal action
    # defined in %setWMax / second.  Resolution is in hundredths of a percent/second.
    # 0 means there is no limit. Interpreted as a % change in output capability limit/second when used as a ramp rate.
    setVNom: Optional[VoltageRMS] = element(default=None)  # AC voltage nominal setting.
    setVRef: Optional[VoltageRMS] = element(
        default=None
    )  # The nominal AC voltage (RMS) at the utility's point of common coupling.
    setVRefOfs: Optional[VoltageRMS] = element(
        default=None
    )  # The nominal AC voltage (RMS) offset between the DER's electrical CP and the utility's point of common coupling.
    updatedTime: types.TimeType = element()  # Specifies the time at which the DER information was last updated.

    # CSIP Aus Extensions (encoded here as it makes decoding a whole lot simpler)
    # This is an encoded version of DOESupportedMode
    doeModesEnabled: Optional[primitive_types.HexBinary8] = element(ns="csipaus", default=None)


class DERListResponse(List, tag="DERList"):
    DER_: Optional[list[DER]] = element(default=None, tag="DER")

    pollRate: Optional[int] = attr(default=types.DEFAULT_POLLRATE_SECONDS)



from pydantic_xml import element

from envoy_schema.server.schema.sep2 import types
from envoy_schema.server.schema.sep2.base import BaseXmlModelWithNS


class FixedVar(BaseXmlModelWithNS):
    """Specifies a signed setpoint for reactive power."""

    refType: types.DERUnitRefType = element()  # Signed setpoint for reactive power
    value: types.SignedPerCent = element()  # Specify a signed setpoint for reactive power in %


class PowerFactorWithExcitation(BaseXmlModelWithNS):
    """Specifies a setpoint for Displacement Power Factor, the ratio between apparent and active powers at the
    fundamental frequency (e.g. 60 Hz) and includes an excitation flag."""

    displacement: int = element()  # Significand of an unsigned value of cos(theta) between 0 and 1.0.
    excitation: bool = element()  # True = DER absorbing, False = DER injecting reactive power
    multiplier: int = element()  # Specifies exponent of 'displacement'.


class PowerFactor(BaseXmlModelWithNS):
    """Specifies a setpoint for Displacement Power Factor, the ratio between apparent and active powers at the
    fundamental frequency (e.g. 60 Hz)."""

    displacement: int = (
        element()
    )  # A setpoint for Displacement PF, the ratio between apparent/active powers at the fundamental frequency eg: 60 Hz
    multiplier: int = element()  # Specifies exponent of 'displacement'.


class ReactivePower(BaseXmlModelWithNS):
    """The reactive power Q"""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in volt-amperes reactive (var) (uom 63)


class ActivePower(BaseXmlModelWithNS):
    """The active/real power P"""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in volt-amperes reactive (var) (uom 63)


class CurrentRMS(BaseXmlModelWithNS):
    """Average flow of charge through a conductor."""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in amperes RMS (uom 5)


class VoltageRMS(BaseXmlModelWithNS):
    """Average electric potential difference between two points."""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in volts RMS (uom 29)


class AmpereHour(BaseXmlModelWithNS):
    """Available electric charge"""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in ampere-hours (uom 106)


class WattHour(BaseXmlModelWithNS):
    """Active (real) energy"""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in watt-hours (uom 72)


class ApparentPower(BaseXmlModelWithNS):
    """The apparent power S (in VA) is the product of root mean square (RMS) voltage and RMS current."""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in volt-amperes (uom 61)


class ReactiveSusceptance(BaseXmlModelWithNS):
    """Reactive susceptance"""

    multiplier: int = element()  # Specifies exponent of 'value'.
    value: int = element()  # Value in siemens (uom 53)



from datetime import datetime
from decimal import Decimal

from pydantic import BaseModel, Field


class BillingReading(BaseModel):
    """Billing readings are a total across all phases for the period"""

    site_id: int
    period_start: datetime
    duration_seconds: int
    value: Decimal  # Positive indicates import, negative indicates export


class BillingTariffRate(BaseModel):
    site_id: int
    period_start: datetime
    duration_seconds: int
    import_active_price: Decimal
    export_active_price: Decimal
    import_reactive_price: Decimal
    export_reactive_price: Decimal


class BillingDoe(BaseModel):
    site_id: int
    period_start: datetime
    duration_seconds: int
    import_limit_active_watts: Decimal
    export_limit_watts: Decimal


class BaseBillingResponse(BaseModel):
    """Billing data common to all billing responses. All readings will come back as a total usage across all phases"""

    tariff_id: int

    varh_readings: list[BillingReading]  # Will be ordered by site_id then period_start
    wh_readings: list[BillingReading]  # Will be ordered by site_id then period_start
    watt_readings: list[BillingReading]  # Will be ordered by site_id then period_start
    active_tariffs: list[BillingTariffRate]  # Will be ordered by site_id then period_start
    active_does: list[BillingDoe]  # Will be ordered by site_id then period_start


class AggregatorBillingResponse(BaseBillingResponse):
    """Response model for a billing report scoped to a particular aggregator for a period of time"""

    aggregator_id: int
    aggregator_name: str
    period_start: datetime
    period_end: datetime


class CalculationLogBillingResponse(BaseBillingResponse):
    """Response model for a billing report scoped to a particular calculation log"""

    calculation_log_id: int


class SiteBillingResponse(BaseBillingResponse):
    """Response model for a billing report scoped to a particular set of site id's for a period of time"""

    site_ids: list[int]
    period_start: datetime
    period_end: datetime


MAX_SITE_IDS_IN_REQUEST = 100


class SiteBillingRequest(BaseModel):
    """Request to send to utility server to request a SiteBillingResponse for the specified parameters.
    (used with uri.SitePeriodBillingUri)"""

    site_ids: list[int] = Field(max_length=MAX_SITE_IDS_IN_REQUEST)  # Limited to MAX_SITE_IDS_IN_REQUEST items
    period_start: datetime
    period_end: datetime
    tariff_id: int



from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2 import base, primitive_types, types


class Resource(base.BaseXmlModelWithNS):
    type: Optional[str] = attr(ns="xsi", default=None)
    href: Optional[str] = attr(default=None)


class IdentifiedObject(Resource):
    mRID: primitive_types.HexBinary128 = element()
    description: Optional[str] = element(default=None)
    version: Optional[types.VersionType] = element(default=None)


class SubscribableResource(Resource):
    subscribable: Optional[types.SubscribableType] = attr(default=None)


class SubscribableList(SubscribableResource):
    """A List to which a Subscription can be requested."""

    all_: int = attr(name="all")  # The number specifying "all" of the items in the list. Required on GET
    results: int = attr()  # Indicates the number of items in this page of results.


class SubscribableIdentifiedObject(SubscribableResource):
    mRID: primitive_types.HexBinary128 = element()  # The global identifier of the object
    description: Optional[str] = element(
        default=None
    )  # The description is a human readable text describing or naming the object.
    version: Optional[types.VersionType] = element(default=None)  # Contains the version number of the object.


class RespondableResource(Resource):
    """A Resource to which a Response can be requested."""

    replyTo: Optional[str] = attr(default=None)
    responseRequired: Optional[primitive_types.HexBinary8] = attr(default=00)


class RespondableSubscribableIdentifiedObject(RespondableResource):
    """An IdentifiedObject to which a Response can be requested."""

    mRID: primitive_types.HexBinary128 = element()
    description: Optional[str] = element(default=None)
    version: Optional[types.VersionType] = element(default=None)
    subscribable: Optional[types.SubscribableType] = attr(default=None)


class List(Resource):
    """Container to hold a collection of object instances or references. See Design Pattern section for additional
    details."""

    all_: int = attr(name="all")  # The number specifying "all" of the items in the list. Required on GET
    results: int = attr()  # Indicates the number of items in this page of results.


class Link(base.BaseXmlModelWithNS):
    href: str = attr()


class ListLink(Link):
    all_: Optional[int] = attr(name="all", default=None)
